#! /usr/bin/env python3
#! -*- coding: utf-8 -*-

import json
import os
import pkg_resources
import re
from taggregator import printer
from pathlib import Path
from json.decoder import JSONDecodeError

class UserConfig:
    """
    The idea of this class is to amalgamate data extracted both from the static config file that
    the user maintains as well as the arguments passed in via the command line at runtime.

    It will also validate that the config file is corrrect and not missing any keys as well as
    creating it if it doesn't exist.

    Its main interface is `config_map` which is a dictionary mapping strings to various types
    of data, for example the runtime root, tags to search for, priorities to search for etc.
    """
    def __init__(self, raw_runtime_args):
        # First we grab everything we can from the config file we find (or create if it doesn't exist)
        self.config_map = get_config_map()

        default_config = get_default_config_json()

        for key in default_config:
            if key not in self.config_map:
                set_config_property(self.config_map, key, default_config[key])

        # We then go on to grab the data from command line arguments that wouldn't have made sense to go in a config file
        self.config_map["root"] = os.path.abspath(raw_runtime_args.root)

        # If we were passed in a set of tags at runtime they take
        # priority and we use them, falling back to tags in config file.
        tags_given_at_runtime = [tag for tag in raw_runtime_args.tags.split(",")] if raw_runtime_args.tags is not None else None
        tags_to_use = tags_given_at_runtime if tags_given_at_runtime is not None else self.config_map["tags"]
        self.config_map["tags"] = set([re.escape(tag.strip().upper()) for tag in tags_to_use])

def get_existing_config_path():
    """
    Look for existing config in first {current dir}/.taggregator and then ~/.taggregator
    """
    current_dir_path = os.path.join(os.getcwd(), os.path.join(CONFIG_FOLDER, CONFIG_FILE_NAME))

    if os.path.isfile(current_dir_path):
        return current_dir_path
    else:
        if os.path.isfile(HOME_DIR_CONFIG_PATH):
            return HOME_DIR_CONFIG_PATH

    return None

def copy_default_config_to_path(path):
    default_config_json = get_default_config_json()
    printer.log("Creating config file at: " + path, "information")

    with open(path, "w") as config_file:
        json.dump(default_config_json, config_file, indent=4)

def get_default_config_json():
    with open(pkg_resources.resource_filename(__name__, "default_config.json"), encoding="utf-8") as default_config_file:
        return json.loads(default_config_file.read())

# @CLEANUP(LOW) get_config_map is a misleading name
# It has the side effect of creating a config file
# if one doesn't exist, but the name implies purity.
def get_config_map():
    """
    Return the content of a config.json if one is found.
    If not, one is created.
    """
    # If neither ~/.taggregator or {current dir}/.taggregator exists,
    # create ~/.taggregator and copy in the default config file from bundle resources
    config_path = get_existing_config_path()

    if not config_path:
        printer.log("No config file found!", "warning")
        config_path = create_default_config_file(HOME_DIR_CONFIG_PATH)

    try:
        with open(config_path) as config_json:
            return json.load(config_json)
    except JSONDecodeError as je:
        error_string = "Error in your taggregator config file at line %d, column %d, exiting..." %(je.lineno, je.colno)
        printer.log(error_string, "fatal error")
        raise SystemExit()

def set_config_property(config, key, value):
    config[key] = value
    config_path = get_existing_config_path()

    printer.log("Found unset config property: '%s', setting it to '%s'" %(key, value), "warning")

    with open(config_path, "w") as config_file:
        json.dump(config, config_file, indent=4)

def create_default_config_file(directory):
    dir_path = os.path.join(directory, ".taggregator")

    if not os.path.exists(dir_path):
        os.makedirs(dir_path)
        printer.log("Creating config directory at: " + dir_path, "information")

    path = os.path.join(dir_path, CONFIG_FILE_NAME)

    if os.path.exists(path):
        while True:
            answer = input("[WARNING] File already exists, do you wish to overwrite it? (y/n): ").lower().strip()
            if answer == "y":
                copy_default_config_to_path(path)
                return path
            elif answer == "n":
                return None

    copy_default_config_to_path(path)
    return path


CONFIG_FOLDER = ".taggregator"
CONFIG_FILE_NAME = "config.json"
HOME_DIR_CONFIG_FOLDER = os.path.join(str(Path.home()), CONFIG_FOLDER)
HOME_DIR_CONFIG_PATH = os.path.join(HOME_DIR_CONFIG_FOLDER, CONFIG_FILE_NAME)



#! /usr/bin/env python3
#! -*- coding: utf-8 -*-

from taggregator import config
from taggregator import printer
from taggregator import tagg
import argparse
import os
import sys
import cProfile, pstats, io

class ProfileData:
    def __init__(self, should_do_profiling, line_count, sort_by):
        self.should_do_profiling = should_do_profiling
        self.line_count = line_count # How many lines the printout should display
        self.sort_by = sort_by

class CommandHandler:
    def __init__(self, profile_data):
        self.profile_data = profile_data

        parser = argparse.ArgumentParser(description="Search files in a directory for areas with specific tags")
        parser.add_argument("command",  help="Command to run")

        allowed_commands = ["run", "create"]
        self.was_run_by_default = len(sys.argv) <= 1 or sys.argv[1] not in allowed_commands
        command = "run" if self.was_run_by_default else sys.argv[1]

        getattr(self, command)()

    def run(self):
        if self.profile_data.should_do_profiling:
            pr = cProfile.Profile()
            pr.enable()

        parser = argparse.ArgumentParser(description="Run the taggregator main program")
        parser.add_argument("root", default="", nargs="?", help="Path from which to start search")
        parser.add_argument("-t", "--tags", help="Comma-separated list of tags to search for (temporarily overrides config file)")

        raw_args = parser.parse_args(sys.argv[1:]) if self.was_run_by_default else parser.parse_args(sys.argv[2:])
        user_config = config.UserConfig(raw_args)

        tagg.run(user_config.config_map)

        if self.profile_data.should_do_profiling:
            pr.disable()
            s = io.StringIO()
            ps = pstats.Stats(pr, stream=s).sort_stats(self.profile_data.sort_by)
            ps.print_stats()

            for i, line in enumerate(s.getvalue().split('\n')):
                if (i < self.profile_data.line_count):
                    print(line)

    def create(self):
        parser = argparse.ArgumentParser(description="Run the taggregator config file creator")
        parser.add_argument("root", default="", nargs="?", help="Directory in which to create config folder")
        raw_args = parser.parse_args(sys.argv[1:]) if self.was_run_by_default else parser.parse_args(sys.argv[2:])

        config.create_default_config_file(raw_args.root)

def main():
    CommandHandler(profile_data)

if __name__ == "__main__":
    CommandHandler(profile_data)

profile_data = ProfileData(False, 30, 'tottime')



#! /usr/bin/env python3
#! -*- coding: utf-8 -*-

from __future__ import print_function # Fix python2 runtime error with end=x
from collections import defaultdict
from os import get_terminal_size
from taggregator.tagg import Match
import math
import statistics

class TerminalColours:
    END             = '\033[0m'
    HEADER          = '\033[1;37m'
    PRIORITY_NONE   = '\033[0m'
    PRIORITY_LOW    = '\033[0;32m'
    PRIORITY_MEDIUM = '\033[0;33m'
    PRIORITY_HIGH   = '\033[0;31m'

def get_truncated_text(text, max_length, truncate_indicator="..."):
    """
    Truncate text at given length and append truncate_indicator if truncated.
    If the text is shorter than or equal to max_length, just return text unmodified
    """
    return (text[:max_length - len(truncate_indicator)] + truncate_indicator) if len(text) > max_length else text

def log(text, tag, append_new_line=False):
    """
    Prints log entries which look like:
    [INFORMATION] Config file created at ~/.taggregator/config.json
    """
    print("[%s] %s" %(tag.upper(), text))

    if (append_new_line):
        print("\n")

def print_right_pad(text, pad_size, append_new_line=False):
    end = "\n" if append_new_line else ""
    print(text + pad_size * " ", end=end)

def print_separator():
    sep_count = int(terminal_columns * 0.75)
    dashes = "-" * sep_count
    print(dashes)

def get_priority_to_colour_map(priority_value_map):
    """
    Map a priority value to a colour based
    on its value relative to the median priority value.
    """
    priority_to_colour_map = { Match.NO_PRIORITY: TerminalColours.PRIORITY_NONE}
    median_value = statistics.median(priority_value_map.values())

    for p in priority_value_map.values():
        if p < median_value:
            priority_to_colour_map[p] = TerminalColours.PRIORITY_LOW
        elif p > median_value:
            priority_to_colour_map[p] = TerminalColours.PRIORITY_HIGH
        else:
            priority_to_colour_map[p] = TerminalColours.PRIORITY_MEDIUM

    return priority_to_colour_map

def get_highlight_colour(orig_colour):
    """
    Keep same colour but switch foreground with background.
    Reference https://en.wikipedia.org/wiki/ANSI_escape_code#SGR_(Select_Graphic_Rendition)_parameters
    """
    return orig_colour.replace("0", "7")

def print_match_line(match, tag_marker, normal_colour, highlighted_colour, pad_size, append_new_line=False):
    """
    Prints a found match, highlighting the tag so that it is clear
    in case there is a line with two different tags.
    """
    line = match.line
    tag = match.tag

    # Find the beginning of the tag and slice the line up into 3 parts:
    # 1) Before the tag 2) During the tag 3) After the tag
    idx = line.find(tag) - len(tag_marker)
    end_idx = idx + len(tag) + len(tag_marker)

    before = line[0:idx]
    during = line[idx:end_idx]
    after = line[end_idx:len(line)]

    # Switch between normal and highlighted based on whether printing the tag
    to_print = normal_colour + before + highlighted_colour + during + normal_colour + after + TerminalColours.END
    print_right_pad(to_print, pad_size, append_new_line)

def print_matches(matches, tag_marker, priority_value_map):
    priority_to_colour_map = get_priority_to_colour_map(priority_value_map)

    # Arrange every match into a dictionary with a key the item's priority,
    # sorted so that we display the highest priority ones at the top.
    matches_by_priority = defaultdict(list)
    matches.sort(key=lambda x: x.priority, reverse=True)

    for match in matches:
        matches_by_priority[match.priority].append(match)

    # Calculate the longest piece of each type of data so that
    # we can do some simple maths to line them up nicely.
    size_longest_name = max([len(match.file_name) for match in matches], default=0)
    size_longest_line_no = max([len(str(match.line_number)) for match in matches], default=0)
    size_longest_line = max([len(match.line) for match in matches], default=0)
    section_padding = 2

    # Looks stupid to draw heading if there are no matches
    if len(matches) > 0:
        print("Your taggregator todo list:")
        print_separator()
    else:
        print("No taggregator tags found - start commenting your code!")

    for p in matches_by_priority: # Grab each key
        matches_by_priority[p].sort(key=lambda x: (x.tag)) # Sort each set of matches by tag in alphabetical order
        for match in matches_by_priority[p]: # Grab each match
            colour = priority_to_colour_map[match.priority]
            highlighted_colour = get_highlight_colour(colour)
            print_right_pad(match.file_name, size_longest_name - len(match.file_name) + section_padding)
            print_right_pad(":" + match.line_number, size_longest_line_no - len(match.line_number) + section_padding)
            print_match_line(match, tag_marker, colour, highlighted_colour, size_longest_line - len(match.line) + section_padding, append_new_line=True)

        # Separator in between sets of matches by priority
        print_separator()

terminal_columns = 80 # Sane default

try:
    terminal_columns = get_terminal_size()[0]
except OSError:
    pass # Leave at 80



#! /usr/bin/env python3
#! -*- coding: utf-8 -*-

from taggregator import printer
from pathlib import Path
import itertools
import os
import re
import sys

class Match:
    NO_PRIORITY = -1

    def __init__(self, file_name, line_number, line, tag, priority):
        self.file_name = file_name
        self.line_number = str(line_number)
        self.line = line
        self.tag = tag
        self.priority = priority

    def __str__(self):
        return self.file_name

    def __eq__(self, other):
        return self.file_name == other.file_name and self.line_number == other.line_number and self.tag == other.tag

def get_piped_list(items):
    return "|".join(items)

def get_tag_regex(tag_marker, tags, priority_regex):
    """
    Get compiled regex for matching tags by:
        -> match tag_marker + tag_string as group
        -> match priority as group
    """

    # @BUG(LOW) Slightly weird matching property
    # Because we have decided that priorities can be optional, we allow zero parentheses around
    # the priority regex. This has the interesting property that the line below would be marked
    # as high priority even though the user might not want it to be:
    # @FEATURE High priority test
    # Not really sure if this is undesired behaviour or not.
    regex_string = tag_marker + "(" + get_piped_list(tags) + ")" + r"\s*\(*" + priority_regex + "\)*"

    # Return regex which will match (for example): @HACK|SPEED|FEATURE(LOW|MEDIUM)
    # with the priority being an optional match
    return re.compile(regex_string, re.IGNORECASE)

def get_priority_regex(priorities):
    return "\s*(" + get_piped_list(priorities) + ")?\s*"

def find_matches(tag_regex, tags, file_name, priority_value_map):
    if os.path.isdir(file_name):
        return

    # @SPEED(HIGH) File opening/reading
    # Profiling shows that this is the greatest bottleneck in the app
    # at the minute, experiments with multiprocessing only slowed it down
    # because it is IO bound work
    with open(file_name) as f:
        # Read whole file into one buffer and see if any of the tags
        # match against it so we dont need to do the expensive regex
        # findall on every line individually unless we find a whole match
        try:
            file_contents = f.read()
        except UnicodeDecodeError:
            # Ignore non utf-8 files
            return

        lower_contents = file_contents.lower()
        lower_tags = [t.lower() for t in tags]

        if not any(t in lower_contents for t in lower_tags):
            return

        # @BUG(HIGH) Throws OSError on some files if in use
        # Can't repro on *nix but happens on Cygwin if the file is in use
        for number, line in enumerate(file_contents.split('\n'), 1):
            # @SPEED(MEDIUM) Regex search of processed line
            matches = tag_regex.findall(line)

            for match in matches:
                tag = match[0].upper()
                priority = match[1]
                priority_idx = priority_value_map.get(priority.upper(), Match.NO_PRIORITY)
                truncated_line = printer.get_truncated_text(line.strip(), 100)

                yield Match(file_name, number, truncated_line, tag, priority_idx)

def get_priority_value_map(all_priorities):
    """
    Maps an index of increasing size to each priority ranging from low -> high
    e.g. given ['LOW', 'MEDIUM', 'HIGH'] will return {'LOW': 0, 'MEDIUM': 1, 'HIGH': 2}
    """
    return dict((priority_text.upper(), priority_index) for priority_index, priority_text in enumerate(all_priorities))

def run(config_map):
    tag_marker = re.escape(config_map["tag_marker"])
    extensions = config_map["extensions"]
    priorities = config_map["priorities"]
    tags = config_map["tags"]

    priority_value_map = get_priority_value_map(priorities)
    value_priority_map = dict(reversed(item) for item in priority_value_map.items())
    priority_regex = get_priority_regex(priorities)
    tag_regex = get_tag_regex(tag_marker, tags, priority_regex)
    exclude = [os.path.join(os.getcwd(), d) for d in config_map["exclude"]]
    can_search_any_extension = "*" in extensions

    files = []

    for root, dirs, files_in_dir in os.walk(os.getcwd()):
        for file_name in files_in_dir:
            file_path = os.path.join(root, file_name)
            # We only want to search for tags in files which have one of the correct
            # extensions (or user has chosen to include every extension with '*')
            # and are not inside one of the excluded folders.
            if can_search_any_extension or any(file_path.endswith(ext) for ext in extensions):
                if not any(file_path.startswith(e) for e in exclude):
                    files.append(file_path)

    matches = []

    for file_name in files:
        for match in find_matches(tag_regex, tags, file_name, priority_value_map):
            # Equality check is handled by the overridden __eq__ in the Match class
            if not any(match == m for m in matches):
                matches.append(match)

    printer.print_matches(matches, tag_marker, priority_value_map)
